// We need this so that we do not get conflicts between Vrui and OSG
#define GLMATRIXTEMPLATES_INCLUDED

#include <Math/Math.h>
#include <Geometry/OrthogonalTransformation.h>
#include <GL/gl.h>
#include <GL/GLColorTemplates.h>
#include <GL/GLVertexTemplates.h>
#include <GL/GLContextData.h>
#include <GL/GLGeometryWrappers.h>
#include <GLMotif/Button.h>
#include <GLMotif/Menu.h>
#include <GLMotif/PopupMenu.h>
#include <Vrui/Vrui.h>
#include <Vrui/Application.h>

#include <AL/ALGeometryWrappers.h>
#include <AL/ALObject.h>
#include <AL/ALContextData.h>

#include <Geometry/Point.h>
#include <Geometry/Vector.h>

#include <Math/Math.h>
#include <Math/Constants.h>

/* Delta3D headers */
#include <dtCore/camera.h>
#include <dtCore/infinitelight.h>
#include <dtCore/transform.h>
#include <dtCore/object.h>
#include <dtCore/scene.h>
#include <dtCore/system.h>
#include <dtCore/environment.h>

/* ODE headers */
#include <ode/ode.h>

/* Delta3D includes */
#include <dtVR/vrapplication.h>

/* osg includes */
#include <osg/Version>
#include <osg/Vec3>
#include <osg/Matrix>
#include <osg/Transform>
#include <osg/Group>
#include <osg/Node>
#include <osg/Camera>
#include <osg/Scissor>
#ifndef DELTA3D
#include <osgDB/ReadFile>
#endif

#include <osgUtil/UpdateVisitor>

#include <osgViewer/Viewer>

//#include <SYNC/MutexPosix.h>
//#include <SYNC/NullMutex.h>

#include <Vrui/ToolInputLayout.h>
#include <Vrui/ToolInputAssignment.h>
#include <Vrui/Vrui.h>
#include "LightSaber.h"
#include "TouchNavigationTool.h"
#include "LeapNavigationTool.h"

#ifdef SHADOW
#include <osgShadow/ShadowedScene>
#include <osgShadow/ShadowVolume>
#include <osgShadow/ShadowTexture>
#include <osgShadow/ShadowMap>
#include <osgShadow/SoftShadowMap>
#include <osgShadow/ParallelSplitShadowMap>
#include <osgShadow/LightSpacePerspectiveShadowMap>
#include <osgShadow/StandardShadowMap>
#endif

#ifdef HIMMEL
#include <osgHimmel/abstractastronomy.h>
#include <osgHimmel/himmel.h>
#include <osgHimmel/timef.h>
#include <osgHimmel/himmelenvmap.h>
#endif

using namespace std;
using namespace dtCore;
using namespace dtABC;
using namespace dtUtil;

#ifdef HIMMEL
using namespace osgHimmel;
#endif

namespace dtCore {
class Environment;
class InfiniteLight;
class Object;
}

class VruiDemo:public Vrui::Application,
#ifdef DELTA3D
	       public VRApplication,
#endif
	       public GLObject, public ALObject
{
  /* Embedded classes: */
private:
  class DataItem:public GLObject::DataItem // Data structure storing OpenGL-dependent application data
  {
    /* Elements: */
  public:
    GLuint textureObjectId; // Texture object ID of some texture
    GLuint displayListId; // Display list ID of some display list
		
    // OSG stuff
    int data;
    osg::Group * root;
    osg::ref_ptr<osgViewer::Viewer> viewer;

    osg::BoundingBox boundingbox;
    float offset[3];
    float scale;

    /* Constructors and destructors: */
    DataItem(void)
    {
      /* Create a texture object to hold a texture: */
      glGenTextures(1,&textureObjectId);
			
      /* Create a display list: */
      displayListId=glGenLists(1);

      offset[0] = offset[1] = offset[2] = 0.0;
    };
    ~DataItem(void)
    {
      /* Destroy the texture object: */
      glDeleteTextures(1,&textureObjectId);
			
      /* Destroy the display list: */
      glDeleteLists(displayListId,1);
    };
  };
	
  class ALDataItem:public ALObject::DataItem // Structure to hold application state associated with a particular OpenAL context
  {
    /* Elements: */
  public:
#if ALSUPPORT_CONFIG_HAVE_OPENAL
    ALuint pulsing, swinging;
    ALuint pulsebuffer, swingbuffer;
#endif
		
    /* Constructors and destructors: */
  public:
    ALDataItem(void)
    {
#if ALSUPPORT_CONFIG_HAVE_OPENAL
      alGenSources(1,&pulsing);
      alGenBuffers(1,&pulsebuffer);
      alGenSources(1,&swinging);
      alGenBuffers(1,&swingbuffer);
#endif
    }
    virtual ~ALDataItem(void)
    {
#if ALSUPPORT_CONFIG_HAVE_OPENAL
      /* Destroy buffers and sources: */
      alDeleteSources(1,&pulsing);
      alDeleteBuffers(1,&pulsebuffer);
      alDeleteSources(1,&swinging);
      alDeleteBuffers(1,&swingbuffer);
#endif
    }

  };

  /* Elements: */
private:
  Vrui::LightSaberTool *lightsaber;

  float scale;
  int frameNumber;
  osg::Node **nodes;
  int noNodes;
  osg::ref_ptr<osg::FrameStamp> frameStamp;
  double lastFrameTime;
#ifdef DELTA3D
  RefPtr<Object> *objects;
#endif
  RefPtr<InfiniteLight> globalInfinite;
  osg::ref_ptr<osg::NodeVisitor> updateVisitor;

#ifdef HIMMEL
  TimeF *g_timef;
  osg::ref_ptr<Himmel> g_himmel;
#endif
  osg::Group *root;

  /* Vrui parameters: */
  GLMotif::PopupMenu* mainMenu; // The program's main menu
	
  /* Private methods: */
  GLMotif::PopupMenu* createMainMenu(void); // Creates the program's main menu
	
  /* Constructors and destructors: */
public:
  VruiDemo(int& argc,char**& argv,char**& appDefaults); // Initializes the Vrui toolkit and the application
  ~VruiDemo(void); // Shuts down the Vrui toolkit

  static ALenum GetFormatFromInfo(short channels, short bitsPerSample) {
    if (channels == 1) {
      if (bitsPerSample == 16)
	return AL_FORMAT_MONO16;
      else
	return AL_FORMAT_MONO8;
    }
    if (bitsPerSample == 16)
      return AL_FORMAT_STEREO16;
    return  AL_FORMAT_STEREO8;
  }

  void setScale (float s) {
    scale = s;
  }

  static void swap (char *c1, char *c2) {
    char c;

    c = *c1;
    *c1 = *c2;
    *c2 = c;
  }

  static int file_read_int32_le (char *buffer, FILE* file) {
    int result;

    fread (buffer, 1, 4, file);
    //swap (buffer, &(buffer[3]));
    //swap (&(buffer[1]), &(buffer[2]));

    result = *((int *)buffer);
    return result;
  }

  static unsigned int file_read_uint32_le (char *buffer, FILE* file) {
    unsigned int result;

    fread (buffer, 1, 4, file);
    //swap (buffer, &(buffer[3]));
    //swap (&(buffer[1]), &(buffer[2]));

    result = *((unsigned int *)buffer);
    return result;
  }

  static short file_read_int16_le (char *buffer, FILE* file) {
    short result;

    fread (buffer, 1, 2, file);
    //swap (buffer, &(buffer[1]));

    result = *((short *)buffer);
    return result;
  }

  static void loadALBuffer(const char *path, ALuint &source, ALuint &buffer)
  {
    int result;

    // Load wav data into a buffers.
    FILE* file = fopen(path, "rb");
    if (!file) {
      printf ("could not open file %s\n", path);
      return;
    }
    char xbuffer[5];
    xbuffer[4] = '\0';
    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "RIFF") != 0)
      printf ("%s is not a WAV file\n", path);

    fread (xbuffer, 1, 4, file);

    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "WAVE") != 0)
      printf ("%s is not a WAV file\n", path);

    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "fmt ") != 0)
      printf ("%s is not a WAV file\n", path);

    file_read_int32_le(xbuffer, file);
    short audioFormat = file_read_int16_le(xbuffer, file);
    short channels = file_read_int16_le(xbuffer, file);
    int sampleRate = file_read_int32_le(xbuffer, file);
    int byteRate = file_read_int32_le(xbuffer, file);
    file_read_int16_le(xbuffer, file);
    short bitsPerSample = file_read_int16_le(xbuffer, file);

#if 0
    if (audioFormat != 16) {
      short extraParams = file_read_int16_le(xbuffer, file);
      for (int i=0; i<extraParams; i++)
	fread (xbuffer, 1, 1, file);
    }
#endif

    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "data") != 0)
      printf ("Invalid WAV file\n");

    int dataChunkSize = file_read_uint32_le(xbuffer, file);
    void *bufferData = malloc (dataChunkSize);
    fread (bufferData, 1, dataChunkSize, file);

    float duration = float(dataChunkSize) / byteRate;
    alBufferData(buffer, GetFormatFromInfo(channels, bitsPerSample), bufferData, dataChunkSize, sampleRate);
    free(bufferData);
    fclose(file);
  }

  const bool isPlaying (ALuint &source) const
  {
    ALenum state;
    
    alGetSourcei(source, AL_SOURCE_STATE, &state);
    
    return (state == AL_PLAYING);
  }

  void createLightSaber ();
  void createTouchTool ();
	
  /* Methods: */
  virtual void initDisplay(GLContextData& contextData) const; // Called once upon creation of each OpenGL context
  virtual void initContext(GLContextData& contextData) const;
  /* Methods from ALObject: */
  virtual void initContext(ALContextData& contextData) const;
  virtual void display(GLContextData& contextData) const; // Called for every eye and every window on every frame
  virtual void sound(ALContextData& contextData) const;
  virtual void frame(void); // Called exactly once per frame
  void resetNavigationCallback(Misc::CallbackData* cbData); // Method to reset the Vrui navigation transformation to its default
};

/*************************
Methods of class VruiDemo:
*************************/

GLMotif::PopupMenu* VruiDemo::createMainMenu(void)
{
  /* Create a popup shell to hold the main menu: */
  GLMotif::PopupMenu* mainMenuPopup=new GLMotif::PopupMenu("MainMenuPopup",Vrui::getWidgetManager());
  mainMenuPopup->setBorderWidth(0.0f);
  mainMenuPopup->setBorderType(GLMotif::Widget::RAISED);
  mainMenuPopup->setBorderColor(Vrui::getUiBgColor());
  mainMenuPopup->setBackgroundColor(Vrui::getUiBgColor());
  mainMenuPopup->setForegroundColor(Vrui::getUiFgColor());
  mainMenuPopup->setMarginWidth(Vrui::getUiSize());
  mainMenuPopup->setTitleSpacing(Vrui::getUiSize());
  mainMenuPopup->setTitle("Vrui OpenSceneGraph",Vrui::getUiFont());
	
  /* Create the main menu itself: */
  GLMotif::Menu* mainMenu=new GLMotif::Menu("MainMenu",mainMenuPopup,false);
  mainMenu->setBorderWidth(0.0f);
  mainMenu->setOrientation(GLMotif::Menu::VERTICAL);
  mainMenu->setNumMinorWidgets(1);
  mainMenu->setMarginWidth(0.0f);
  mainMenu->setSpacing(Vrui::getUiSize());
	
  /* Create a button: */
  GLMotif::Button* resetNavigationButton=new GLMotif::Button("ResetNavigationButton",mainMenu,"Reset Navigation",Vrui::getUiFont());
	
  /* Add a callback to the button: */
  resetNavigationButton->getSelectCallbacks().add(this,&VruiDemo::resetNavigationCallback);
	
  /* Finish building the main menu: */
  mainMenu->manageChild();
	
  return mainMenuPopup;
}

VruiDemo::VruiDemo(int& argc,char**& argv,char**& appDefaults)
  :Vrui::Application(argc,argv,appDefaults),
#ifdef DELTA3D
   VRApplication(true),
#endif
   mainMenu(0)
{
  /* Request sound processing: */
  Vrui::requestSound();

  int i, start;
  scale = 1.0;

  /* Create the user interface: */
  mainMenu=createMainMenu();
	
  /* Install the main menu: */
  Vrui::setMainMenu(mainMenu);
	
  /* Set the navigation transformation: */
  resetNavigationCallback(0);

  if (argc > 1 && argv[1][0] == '-' && argv[1][1] == 's') {
    noNodes = argc - 2;
    start = 2;
    sscanf (&(argv[1][2]), "%f", &scale);
    printf ("scale: %f\n", scale);
  } else {
    noNodes = argc - 1;
    start = 1;
  }
  nodes = new osg::Node*[noNodes];
#ifdef DELTA3D
  objects = new RefPtr<Object>[noNodes];
#endif
  for (i=start; i<argc; i++) {
#ifdef DELTA3D
    objects[i-start] = new Object("Object");
    nodes[i-start] = objects[i-start]->LoadFile(argv[i]);
#else
    nodes[i-start] = osgDB::readNodeFile(argv[i]);
#endif
  }

#if DELTA3D
  // create an infinite light
  globalInfinite = new InfiniteLight(4, "GlobalInfiniteLight");
  globalInfinite->SetTransform (Transform (0.0, 0.0, 0.0, 0.0, 0.0, 90.0));
  GetScene()->AddDrawable(globalInfinite.get());
  globalInfinite->SetEnabled(true);
  globalInfinite = new InfiniteLight(5, "GlobalInfiniteLight");
  globalInfinite->SetTransform (Transform (0.0, 0.0, 0.0, 0.0, 90.0, 0.0));
  GetScene()->AddDrawable(globalInfinite.get());
  globalInfinite->SetEnabled(true);
  globalInfinite = new InfiniteLight(6, "GlobalInfiniteLight");
  globalInfinite->SetTransform (Transform (0.0, 0.0, 0.0, 90.0, 0.0, 0.0));
  GetScene()->AddDrawable(globalInfinite.get());
  globalInfinite->SetEnabled(true);

  for (i=0; i<noNodes; i++)
    GetScene()->AddDrawable(objects[i].get());
#endif

  // Tell osg::Referenced to use thread-safe reference counting. We do
  // this here for all applications and all configurations though it is
  // only strictly necessary when Vrui is configured to use
  // multi-threaded rendering.
  osg::Referenced::setThreadSafeReferenceCounting(false);

  /* Initialize update visitor */
  updateVisitor = new osgUtil::UpdateVisitor();
  frameStamp = new ::osg::FrameStamp();
  updateVisitor->setFrameStamp(frameStamp.get());

#ifdef HIMMEL
  g_timef = new TimeF(time(NULL), - 3600.0L * 2.0L, 3600.0L);
  g_himmel = Himmel::createWithoutClouds();
#endif

  root = new osg::Group();
#ifdef SHADOW
  osgShadow::ShadowedScene *scene = new osgShadow::ShadowedScene;
  osg::ref_ptr<osgShadow::SoftShadowMap> sm = new osgShadow::SoftShadowMap;
  scene->setShadowTechnique(sm.get());
#else
  osg::Group *scene = new osg::Group();
#endif
  root->setName("Root");
  root->getOrCreateStateSet ()->
    setMode (GL_CULL_FACE,
	     osg::StateAttribute::OFF | osg::StateAttribute::OVERRIDE);

#if 0
  osg::Light * light = new osg::Light();
  light->setAmbient(osg::Vec4d(0.4, 0.4, 0.4, 1.0));
  light->setDiffuse(osg::Vec4d(0.3, 0.3, 0.3, 1.0));
  light->setSpecular(osg::Vec4d(0.3, 0.3, 0.3, 1.0));
  light->setPosition(osg::Vec4d(0.0, 0.0, 0.0, 1.0));
  light->setConstantAttenuation (0.3);
  light->setLinearAttenuation (0.0);
  light->setQuadraticAttenuation (0.0);
  light->setLightNum (1);
  osg::LightSource *lightsource = new osg::LightSource();
  lightsource->setLight(light);
  osg::StateSet *stateset = root->getOrCreateStateSet();
  lightsource->setStateSetModes(*stateset, osg::StateAttribute::ON);
  root->addChild (lightsource);
#endif

  root->addChild (scene);
  for (int i=0; i<noNodes; i++)
    scene->addChild(nodes[i]);

#ifdef HIMMEL
  g_timef->start();

  g_himmel->assignTime(g_timef);
  g_himmel->setViewSizeHint(1920, 1080);

  g_himmel->setAltitude(100.043);
  //g_himmel->setLatitude(39.7505);
  //g_himmel->setLongitude(84.268);
  g_himmel->setLatitude(52.5491);
  g_himmel->setLongitude(13.3611);

  root->addChild(g_himmel);
#endif
}

VruiDemo::~VruiDemo(void)
{
  delete mainMenu;
}

void VruiDemo::createLightSaber ()
{
  Vrui::LightSaberToolFactory *toolfactory;
  Vrui::ToolInputLayout layout;
  layout.setNumButtons (1);
  Vrui::ToolInputAssignment assignment (layout);
  Vrui::ToolManager *toolmanager = Vrui::getToolManager ();

  toolfactory =
    new Vrui::LightSaberToolFactory (*toolmanager);

  toolmanager->addClass (toolfactory);
}

void VruiDemo::initDisplay(GLContextData& contextData) const
{
  /*********************************************************************
	This method is called once for every OpenGL context created for the
	program. It must not change application state, but only create
	OpenGL-dependent application data and store them in the GLContextData
	object for retrieval in the display method.
  *********************************************************************/

  glDepthFunc(GL_LESS);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glEnable (GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
}
	
void VruiDemo::initContext(GLContextData& contextData) const
{
  /* Create a new context data item: */
  DataItem* dataItem = new DataItem();

  // Create the osgViewer instance
  osg::ref_ptr<osgViewer::Viewer> viewer = new osgViewer::Viewer;
  viewer->setThreadingModel(osgViewer::Viewer::SingleThreaded);

  // Set up osgViewer::GraphicsWindowEmbedded for this context
  osg::ref_ptr<osg::GraphicsContext::Traits> traits =
    new ::osg::GraphicsContext::Traits;
  osg::ref_ptr<osgViewer::GraphicsWindowEmbedded> graphicsWindow =
    new osgViewer::GraphicsWindowEmbedded(traits.get());
  viewer->getCamera()->setGraphicsContext(graphicsWindow.get());
  viewer->getCamera()->getGraphicsContext()->getState()->setContextID(
								      osg::GraphicsContext::createNewContextID());
  viewer->getCamera()->getView()->setLightingMode(osg::View::NO_LIGHT);
  viewer->setLightingMode(osg::View::NO_LIGHT);

  viewer->getCamera()->setComputeNearFarMode(osgUtil::CullVisitor::DO_NOT_COMPUTE_NEAR_FAR);

  viewer->getCamera()->setClearColor(::osg::Vec4(0.0f, 0.0f, 0.0f, 0.0f));
  viewer->getCamera()->setClearMask(0);

#ifdef HIMMEL
  g_himmel->setCameraHint(viewer->getCamera());
#endif

  osg::BoundingSphere boundingSphere = root->computeBound ();
  osg::Vec3f vector = boundingSphere.center ();
  dataItem->scale = scale * 120.0 / boundingSphere.radius ();
  dataItem->offset[0] = -vector[0];
  dataItem->offset[1] = -vector[1];
  dataItem->offset[2] = -vector[2];

  dataItem->root = root;

  // Add the tree to the viewer and set properties
  //Guard<MutexPosix> viewerGuard(dataItem->viewerLock);
  viewer->setSceneData(root);

  dataItem->viewer = viewer;

  contextData.addDataItem(this, dataItem);
}

void VruiDemo::initContext(ALContextData& contextData) const
{
  /* Create a new context data item: */
  ALDataItem* dataItem=new ALDataItem;
  contextData.addDataItem(this, dataItem);

#if ALSUPPORT_CONFIG_HAVE_OPENAL
  /* Generate buffers and sources: */
  loadALBuffer ("sounds/lightsaberpulse.wav",
		dataItem->pulsing,
		dataItem->pulsebuffer);

  alSourceBuffer(dataItem->pulsing, dataItem->pulsebuffer);
  alSourceLooping(dataItem->pulsing, AL_TRUE);
  alSourcePitch(dataItem->pulsing, 1.0f);
  alSourceGain(dataItem->pulsing, 17.0f);

  loadALBuffer ("sounds/ltsaberswing06.wav",
		dataItem->swinging,
		dataItem->swingbuffer);

  alSourceBuffer(dataItem->swinging, dataItem->swingbuffer);
  alSourceLooping(dataItem->swinging, AL_FALSE);
  alSourcePitch(dataItem->swinging, 1.0f);
  alSourceGain(dataItem->swinging, 17.0f);
#endif
}

void VruiDemo::display(GLContextData& contextData) const
{
  /*********************************************************************
	This method is called once for every eye in every window on every
	frame. It must not change application state, as it is called an
	unspecified number of times. It also must not clear the screen or
	initialize the OpenGL transformation matrices. When this method is
	called, Vrui will already have rendered its own state (menus etc.) and
	have set up the transformation matrices so that all rendering in this
	method happens in navigation (i.e., model) coordinates.
  *********************************************************************/
	
  /* Get the OpenGL-dependent application data from the GLContextData object: */
  DataItem* dataItem=contextData.retrieveDataItem<DataItem>(this);
	
  // normalize it so that we can fully see the scene
  glPushAttrib( GL_VIEWPORT_BIT );
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glMatrixMode( GL_TEXTURE );
  glPushMatrix();
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  //glRotatef (90.0, 1.0, 0.0, 0.0);
#if 1
  glScalef (dataItem->scale, dataItem->scale, dataItem->scale);
  glTranslatef (dataItem->offset[0], dataItem->offset[1], dataItem->offset[2]);
#endif

  dataItem->viewer->advance(lastFrameTime);
  if (!dataItem->viewer->done()) {
    dataItem->viewer->updateTraversal();
  }

#if 0
  GLfloat light_ambient[] = { 0.8, 0.8, 0.8, 1.0 };
  GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat light_position[] = { 0.0, 0.0, 0.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT1, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT1, GL_POSITION, light_position);
  glEnable (GL_LIGHT1);
#endif

  GLint vp[4];
  glGetIntegerv(GL_VIEWPORT, vp);
  GLdouble mv[16], p[16];
  glGetDoublev(GL_MODELVIEW_MATRIX, mv);
  glGetDoublev(GL_PROJECTION_MATRIX, p);

  dataItem->viewer->getCamera()->setViewport(vp[0], vp[1], vp[2], vp[3]);
  dataItem->viewer->getCamera()->setProjectionMatrix(osg::Matrix(p));
  dataItem->viewer->getCamera()->setViewMatrix(osg::Matrix(mv));

  /* Render all opaque surfaces: */
  glDisable (GL_CULL_FACE);
  dataItem->viewer->renderingTraversals();
  glEnable (GL_CULL_FACE);

  // since OSG turns scissor mode off we need to turn it back on just
  // in case we render stereo as left/right
  glEnable (GL_SCISSOR_TEST);

  glMatrixMode( GL_MODELVIEW );
  glPopMatrix();
  glMatrixMode( GL_TEXTURE );
  glPopMatrix();
  glMatrixMode( GL_PROJECTION );
  glPopMatrix();
  glPopAttrib();
  glMatrixMode( GL_MODELVIEW );
}

void VruiDemo::sound (ALContextData& contextData) const
{
  typedef ALContextData::Point ALPoint;
  typedef ALContextData::Transform ALTransform;

  /* Get the data item: */
  ALDataItem* dataItem=contextData.retrieveDataItem<ALDataItem>(this);
	
  /* Get the current modelview matrix from the OpenAL context: */
  const ALTransform& transform=contextData.getMatrix();
	
#if ALSUPPORT_CONFIG_HAVE_OPENAL
  if (lightsaber && lightsaber->isActive ()) {
    /* Set the source position transformed to physical coordinates: */
    alSourcePosition(dataItem->pulsing,
		     ALPoint(lightsaber->getCurrentOrigin ()),
		     transform);

    /* Set the source attenuation factors in physical coordinates: */
    alSourceReferenceDistance(dataItem->pulsing,Vrui::getDisplaySize());
    alSourceRolloffFactor(dataItem->pulsing,1.0);

    if (!isPlaying (dataItem->pulsing))
      alSourcePlay(dataItem->pulsing);

    // check if sufficient movement for swinging noise
    static Geometry::Point<float,3> *oldorigin = 0;
    static Geometry::Vector<float,3> oldaxis;
    if (!oldorigin) {
      oldorigin = new Geometry::Point<float,3>;
      *oldorigin = lightsaber->getCurrentOrigin ();
      oldaxis = lightsaber->getCurrentAxis ();
    }
    else {
      Geometry::Point<float,3> origin = lightsaber->getCurrentOrigin ();
      Geometry::Vector<float,3> axis = lightsaber->getCurrentAxis ();

      if (Geometry::sqrDist (origin, *oldorigin) > 1.0e-1 ||
	  fabs (axis[0]*oldaxis[0] + axis[1]*oldaxis[1] + axis[2]*oldaxis[2]) < 0.95) {
	if (!isPlaying (dataItem->swinging))
	  alSourcePlay(dataItem->swinging);
      }

      *oldorigin = origin;
      oldaxis = axis;
    }
  }
  else {
    alSourceStop(dataItem->pulsing);
  }
#endif
}

void VruiDemo::frame(void)
{
  /*********************************************************************
	This function is called exactly once per frame, no matter how many
	eyes or windows exist. It is the perfect place to change application
	state (run simulations, animate models, etc.).
  *********************************************************************/
	
  /* Get the time since the last frame: */
  double frameTime=Vrui::getApplicationTime();

#ifdef HIMMEL	
  // update time for sky rendering
  g_timef->update ();
#endif

  ++frameNumber;

  // Update the frame stamp with information from this frame.
  frameStamp->setFrameNumber(frameNumber);
  frameStamp->setReferenceTime(frameTime);
  frameStamp->setSimulationTime(frameTime);

  // Set up the time and frame number so time-dependent things (animations,
  // particle system) function correctly.
  updateVisitor->setTraversalNumber(frameNumber);

  lastFrameTime = frameTime;

  lightsaber = Vrui::LightSaberTool::getTool ();

  //Vrui::requestUpdate ();
}

void VruiDemo::resetNavigationCallback(Misc::CallbackData* cbData)
{
  /* Reset the Vrui navigation transformation: */
  Vrui::NavTransform t=Vrui::NavTransform::identity;
  t*=Vrui::NavTransform::translateFromOriginTo(Vrui::getDisplayCenter());
  t*=Vrui::NavTransform::scale(Vrui::getInchFactor());
  Vrui::setNavigationTransformation(t);
	
  /*********************************************************************
	Now the coordinate system's origin is in the middle of the
	environment, e.g., in the middle of the workbench, and one coordinate
	unit is one inch long.
  *********************************************************************/
}

int main(int argc,char* argv[])
{
  /* Create an application object: */
  char** appDefaults=0; // This is an additional parameter no one ever uses
  VruiDemo app(argc,argv,appDefaults);
  //for (int i=0; i<argc; i++)
    //printf ("file[%d] %s\n", i, argv[i]);

  app.createLightSaber ();
  Vrui::TouchNavigationTool::createTouchTool ();
  Vrui::LeapNavigationTool::createLeapTool ();

  /* Run the Vrui main loop: */
  app.run();
	
  /* Exit to OS: */
  return 0;
}
